var noteInfo = null;
var serviceName = "Evernote";
var localeStrings = {
  enfPopupForward: "",
  enfPopupTitle: "",
  enfPopupAppend: "",
  enfPopupNotebook: "",
  enfPopupReminder: "",
  enfPopupNotification: "",
  enfPopupTags: ""
};

window.addEventListener("unload",()=>{
  let obj = {};
  obj["height"] = window.outerHeight;
  obj["width"] = window.outerWidth;
  //Directly call background script function to save window size as prefs.
  //If sendMessage() is used to communicate with the background script or 
  //is used browser.storage.local.set(), promise will be resolved after this context is unloaded.
  let bgpage = browser.extension.getBackgroundPage();
  bgpage.savePrefs(obj);
});

async function init() {
  document.getElementById("fwdButton").addEventListener("click", (event) => {
    sendResult(true);
  });
  document.getElementById("cancelButton").addEventListener("click", (event) => {
    sendResult(false);
  });

  document.getElementById("yesCancelButton").addEventListener("click", (event) => {
    sendCancel(true);
  });
  document.getElementById("noCancelButton").addEventListener("click", (event) => {
    sendCancel(false);
  });

  browser.runtime.onMessage.addListener(async (message) => {
    switch (message.message) {
      case "SEND_NOTEINFO":
        serviceName = message.serviceName;
        if (message.progress === 2) {
          noteInfo = message.noteInfo;
          fillNotePanels(message.email && message.email.indexOf("@") > 0);
        } else if (message.progress === 3) {
          showCancelForm();
        }
        break;
      default:
        break;
    }
  });

  browser.runtime.sendMessage({
    message: "START_FORWARDING"
  });
  
  translate();
  for (let key in localeStrings) {
    localeStrings[key] = browser.i18n.getMessage(key);
  }
}

init();

function showCancelForm() {
  let caption = document.getElementById("serviceName");
  caption.textContent = caption.textContent.replace("%s", serviceName);
  document.getElementById("confArea").setAttribute("class", "box-none");
  document.getElementById("cancelArea").setAttribute("class", "box-visible");
}

async function fillNotePanels(haveEmail) {
  let tabsPanel = null;
  
  if (noteInfo[0].onenote) {
    document.getElementById("entable").setAttribute("class", "box-none");
    document.getElementById("ontable").setAttribute("class", "box-visible");
    tabsPanel = document.getElementById("ontabs");
  } else {
    document.getElementById("ontable").setAttribute("class", "box-none");
    document.getElementById("entable").setAttribute("class", "box-visible");
    tabsPanel = document.getElementById("entabs");
  }
  
  let caption = document.getElementById("serviceName");
  caption.textContent = caption.textContent.replace("%s", serviceName);

  for (let i = 0; i < noteInfo.length; i++) {
    addTabInput(tabsPanel, i + 1);
  }
  //hide tab index if only one message is forwarded
  if (noteInfo.length == 1) {
    document.getElementById("tabLabel1").classList.add("box-none");
  }

  for (let i = 0; i < noteInfo.length; i++) {
    let info = noteInfo[i];
    let tabContent = createNotePanel(info, i);
    tabsPanel.appendChild(tabContent);
  }

  addTabCSS(noteInfo.length);
  document.getElementById("tab1").checked = true;

  if (haveEmail) {
    document.getElementById("noemail_msgs").setAttribute("class", "box-none");
    document.getElementById("tableArea").setAttribute("class", "box-visible");
    document.getElementById("fwdButton").setAttribute("class", "box-visible");
  } else {
    document.getElementById("noemail_msgs").setAttribute("class", "box-visible");
    document.getElementById("tableArea").setAttribute("class", "box-none");
    document.getElementById("fwdButton").setAttribute("class", "box-none");
  }
}

function addTabCSS(num) {
  let items = [];
  for (let i=0; i<num; i++) {
    let tab = "#tab" + (i + 1);
    let str = tab + ":checked ~ " + tab + "_content";
    items.push(str);
  }

  let cssStr = items.join(",");
  cssStr += "{display: block;}";
  document.styleSheets[0].insertRule(cssStr);
}

function addTabInput(panel, index) {
  let input = document.createElement("input");
  input.setAttribute("type", "radio");
  input.setAttribute("name", "tab_item");
  input.setAttribute("id", "tab" + index);
  panel.appendChild(input);

  let label = document.createElement("label");
  label.setAttribute("class", "tab_item");
  label.setAttribute("for", "tab" + index);
  label.setAttribute("id", "tabLabel" + index);
  label.textContent = index;
  panel.appendChild(label);
}

function createNotePanel(info, num) {
  let index = num + 1;
  let tabContent = document.createElement("div");
  tabContent.setAttribute("class", "tab_content");
  tabContent.setAttribute("id", "tab" + index + "_content");

  //Forward
  let label = document.createElement("label");
  label.textContent = localeStrings.enfPopupForward + ": ";
  label.style.fontWeight = "bold";
  let elem = document.createElement("input");
  elem.setAttribute("type", "checkbox");
  elem.checked = !info.canceled;
  elem.setAttribute("id", "fwd" + num);
  let p = createNotePanelItem(label, elem);
  tabContent.appendChild(p);

  //Title
  label = document.createElement("label");
  label.textContent = localeStrings.enfPopupTitle + ":";
  label.style.fontWeight = "bold";
  elem = document.createElement("input");
  elem.setAttribute("type", "text");
  elem.setAttribute("class", "tab_textinput");
  elem.setAttribute("value", info.title);
  elem.setAttribute("id", "title" + num);
  p = createNotePanelItem(label, elem, true);
  tabContent.appendChild(p);

  if (!info.onenote) {
    //Append
    label = document.createElement("label");
    label.textContent = localeStrings.enfPopupAppend + ": ";
    label.style.fontWeight = "bold";
    elem = document.createElement("input");
    elem.setAttribute("type", "checkbox");
    elem.checked = info.append;
    elem.setAttribute("id", "append" + num);
    p = createNotePanelItem(label, elem);
    tabContent.appendChild(p);

    //Notebook
    label = document.createElement("label");
    label.textContent = localeStrings.enfPopupNotebook + ":";
    label.style.fontWeight = "bold";
    elem = document.createElement("input");
    elem.setAttribute("type", "text");
    elem.setAttribute("value", info.notebook);
    elem.setAttribute("id", "notebook" + num);
    elem.setAttribute("class", "tab_textinput");
    p = createNotePanelItem(label, elem, true);
    tabContent.appendChild(p);

    //Reminder
    label = document.createElement("label");
    label.textContent = localeStrings.enfPopupReminder + ": ";
    label.style.fontWeight = "bold";
    elem = document.createElement("input");
    elem.setAttribute("type", "checkbox");
    elem.checked = info.reminder;
    elem.setAttribute("id", "reminder" + num);
    p = createNotePanelItem(label, elem);
    tabContent.appendChild(p);

    //Notification (Reminder date)
    label = document.createElement("label");
    label.textContent = localeStrings.enfPopupNotification + ":";
    label.style.fontWeight = "bold";
    elem = document.createElement("input");
    elem.setAttribute("type", "text");
    elem.setAttribute("value", info.reminderDate);
    elem.setAttribute("placeholder", "yyyy/mm/dd");
    elem.setAttribute("id", "reminderdate" + num);
    elem.setAttribute("class", "tab_textinput");
    p = createNotePanelItem(label, elem, true);
    tabContent.appendChild(p);
    
    //Tags
    label = document.createElement("label");
    label.textContent = localeStrings.enfPopupTags + ":";
    label.style.fontWeight = "bold";
    elem = document.createElement("input");
    elem.setAttribute("type", "text");
    elem.setAttribute("value", info.tags.join(", "));
    elem.setAttribute("placeholder", "Tag1, Tag2, ...");
    elem.setAttribute("id", "tags" + num);
    elem.setAttribute("class", "tab_textinput");
    p = createNotePanelItem(label, elem, true);
    tabContent.appendChild(p);
  }

  return tabContent;
}

function createNotePanelItem(label, elem, needBR) {
  let p = document.createElement("p");
  p.appendChild(label);
  needBR = true;
  if (needBR) {
    let br = document.createElement("br");
    p.appendChild(br);
    elem.style.marginLeft = "20px";
  }
  p.appendChild(elem);

  return p;
}

function applyUserChanges(confirmed) {
  let sendNum = 0;
  if (noteInfo) {
    for (let i = 0; i < noteInfo.length; i++) {
      let info = noteInfo[i];
      let elem = document.getElementById("fwd" + i);
      info.canceled = confirmed ? !elem.checked : true;
      elem = document.getElementById("title" + i);
      info.title = elem.value;
      if (!info.onenote) {
        elem = document.getElementById("append" + i);
        info.append = elem.checked;
        elem = document.getElementById("notebook" + i);
        info.notebook = elem.value;
        elem = document.getElementById("reminder" + i);
        info.reminder = elem.checked;
        elem = document.getElementById("reminderdate" + i);
        info.reminderDate = elem.value;
        elem = document.getElementById("tags" + i);
        info.tags = elem.value ? elem.value.split(/\s*,\s*/) : [];
      }

      if (!info.canceled) {
        sendNum++;
      }
    }
  }

  return sendNum;
}

async function sendResult(confirmed) {
  let sendNum = applyUserChanges(confirmed);

  await browser.runtime.sendMessage({
    message: "USER_CHECKED",
    confirmed: confirmed,
    noteInfo: noteInfo,
    sendNum: sendNum
  });

  //close popup automatically
  window.close();
}

async function sendCancel(canceled) {
  if (canceled) {
    await browser.runtime.sendMessage({
      message: "USER_CANCELED"
    });
  }

  //close popup automatically
  window.close();
}
